#include <v_lib.h>
#include "framebuf.h"
#include "cspace.h"

/*
typedef struct dtree_node_s
{
	char	id[MAX_ID];
	int		type; // matrix, vector, output
	vboolean evaluated;
	union	operand
	{
		matrix_p matrix;
		vector_p vector;
		framebuf_p framebuf;
	} op;
	struct	dtree_node_s *parent; // NULL if type==output
	u32_t	num_children;
	struct	dtree_node_s *children[MAX_DEPS];
}

	

random_seed 3

matrix matrix1 # name - any string
{
	# state parameter; reset after evaluation
	load_width		256
	# load_width width_of img1.bmp
	# load_width power2
	# load_height power2
	# load_size power2
	# load_size power2_square
	# load_size size_of matrix2
	# state parameter
	load_height		256
	# operand
	source			image img1.bmp channel red # green,blue,alpha,hue,saturation,value,cieX,cieY,cieZ
	# source		constant 0.5
	# source		random 0.1 0.3
	# state stack empty, result: framebuffer in cache, red channel extracted to matrix
	fourier_forward
	# result: 2D Fourier transform
	# multiply constant 0.35
	# multiply matrix1 - by-element multiplication
	# add matrix2
	zero_rows		1 2
	zero_rows		15 255
	zero_cols		1 2
	zero_cols		15 255
	fourier_inverse
	unify
	store		image1_temp
	gamma_correction vector2 # see below for vector2
	destination		out.bmp:red
}

vector vector2
{
	historgam	image1_temp
	save img1_red.txt # optional
	integrate
}

output out.tga
{
	source_space RGB	# or HSV or CIE
	red 	matrix1
	green	matrix2
	blue	matrix3
	alpha	matrix_05	# allowed for any colorspace
}

processing:
	1. Construct dependency tree (checking for circular refs)
	2. Evaluate from leafs recursively
	3. Destroy from leafs recursively
*/
#if 0

/* ==================================================================== */


void zero_freq(vec_t *afreq,s32_t lo_pass,s32_t hi_pass)
{
	u32_t hlen;
	/*
	 * Discard frequencies below 'lo_pass'
	 */
	if(lo_pass)
		memZero(afreq,lo_pass * 2 * sizeof(vec_t));
	/*
	 * Discard frequencies above 'hi_pass'
	 */
	hi_pass <<= 1;
	hlen = SCAN_WIDTH - hi_pass;
	if(hlen)
		memZero(afreq + hi_pass,hlen * sizeof(vec_t));
}

void zero2d(matrix_t *m,s32_t lo_pass,s32_t hi_pass)
{
	u32_t i;
	matrix_t *m2 = matAlloc(m->rows,m->cols);
	for(i=0;i<m->rows;i++)
		zero_freq(m->cell[i],lo_pass,hi_pass);
	matTranspose(m2,m);
	for(i=0;i<m->rows;i++)
		zero_freq(m2->cell[i],lo_pass,hi_pass);
	matTranspose(m,m2);
	memFree(m2);
}

vhandle_t fft_context = LIBV_INVALID_HANDLE;

void band_pass(u8_p pixels,s32_t lo_pass,s32_t hi_pass)
{
	u32_t channel;
	for(channel=0;channel<3;channel++)
	{
		scanlines_to_array(pixels,channel);
		fftForward2D(fft_context,&pixmat);
		zero2d(&pixmat,lo_pass,hi_pass);
		fftInverse2D(fft_context,&pixmat);
		array_to_scanlines(pixels,channel);
	}
}

void masked_zero(matrix_t *m,matrix_t *mask)
{
	u32_t i,j;
	for(i=0;i<m->rows;i++)
		for(j=0;j<m->cols;j++)
			m->cell[i][j] *= mask->cell[i][j];
}

void mask_pass(u8_p pixels,matrix_t *mask)
{
	u32_t channel;
	for(channel=0;channel<3;channel++)
	{
		scanlines_to_array(pixels,channel);
		fftForward2D(fft_context,&pixmat);
		masked_zero(&pixmat,mask);
		fftInverse2D(fft_context,&pixmat);
		array_to_scanlines(pixels,channel);
	}
}

/* ==================================================================== */

void usage(char *name)
{
	puts("Band-pass image filtering by FDFT/IDFT. Version 0.2 by Vladimir A. Smirnov.\n");
	printf("Usage: %s -infile <infile> [-space-hsv -desaturate]"
		"(-bounds <low bound> <high bound> | -mask mask <maskfile>) "
		"[-outfile outfile] [-outmethod <outmethod>], where:\n",name);
	puts("\t'infile' - name of image file to process (BMP, TGA or JPG);");
	puts("\t'low bound', 'high bound' - bounds in wave-number domain;");
	puts("\t'outfile' - optional name of output file (default is 'out.bmp').");
	printf("Width of the input bitmap must be %u, height must not be greater than %u.\n",SCAN_WIDTH,SCAN_WIDTH);
	printf("Bounds in wave-number space must be in range [0..%u].\n",SCAN_WIDTH>>1);
	puts("'outmethod' must be either 'clamp' (clamp results of IDFT in range 0..255),");
	puts("or 'compensate' (increase contrast such that luma occupy range from 0 to 255),");
	puts("or 'quant' (same as compensate, but produce mono output); default is 'clamp'.");
	puts("If 'outmethod' is 'quant', it's possible to specify threshold (ex.: quant=77).");
	puts("KNOWN BUGS: Probably, a lot.");
}

vboolean
init(int argc,char *argv[],
	s32_t *l_ret,s32_t *h_ret,char **inname_ret,char **outname_ret)
{
	int i;
	s32_t fmax = SCAN_WIDTH >> 1;
	init_pixmat();
	*outname_ret = NULL;
	*inname_ret = NULL;
	for(i=1;i<argc;)
	{
		char *token = argv[i++];
		if(!strcasecmp(token,"-infile"))
		{
			if(i >= argc)
				return false;
			*inname_ret = argv[i++];
			continue;
		} else if(!strcasecmp(token,"-outfile"))
		{
			if(i >= argc)
				return false;
			*outname_ret = argv[i++];
			continue;
		} else if(!strcasecmp(token,"-bounds"))
		{
			if((i+1) >= argc)
				return false;
    		if(!strToInt(argv[i++],l_ret) || !strToInt(argv[i++],h_ret))
	    	{
    			conlPuts(LOG_ERROR,
    				"init(): expected two integers after -bounds.\n");
    			return false;
    		}
    		if(*l_ret < 0 || *l_ret > fmax || *h_ret < 0 || *h_ret > fmax)
	    	{
    			conlPrintf(LOG_ERROR,
    				"init(): wrong frequency bounds '%d' and '%d'.\n"
    				"For help, run %s without any arguments.\n",
    				*l_ret,*h_ret,argv[0]);
    			return false;
    		}
			continue;
		} else if(!strcasecmp(token,"-mask"))
		{
			if(i >= argc)
				return false;
	    	mask = LoadMask(argv[i++]);
    		if(mask == NULL)
    			return false;
		} else if(!strcasecmp(token,"-outmethod"))
		{
			if(i >= argc)
				return false;
			if(!set_clamp(argv[i++]))
			{
				conlPuts(LOG_ERROR,"init(): wrong outmethod.\n");
				return false;
			}
			continue;
		} else if(!strcasecmp(token,"-space-hsv"))
		{
			space_hsv = true;
			continue;
		} else if(!strcasecmp(token,"-desaturate"))
		{
			desaturate = true;
			continue;
		}
	}
	fft_context = fftCreateContext(SCAN_POWER,true);
	return true;
}

int old_main(int argc,char *argv[])
{
	s32_t lo_pass,hi_pass;
	u8_t *pixels;
	char *inname, *outname;
	logStartConsoleLog();
	logUpperLevel(LOG_TRACE);
	if(!init(argc,argv,&lo_pass,&hi_pass,&inname,&outname) ||
		!inname)
	{
		usage(argv[0]);
		return -1;
	}
	if(!outname)
		outname = "out.bmp";
	pixels = LoadImageEx(inname,&image_width,&image_height,space_hsv);
	if(!pixels)
	{
		if(mask)
			memFree(mask);
		return -1;
	}
	if(image_width != SCAN_WIDTH)
	{
		conlPrintf(LOG_ERROR,"Bitmap width must be equal to %u, got %u\n",SCAN_WIDTH,image_width);
		memFree(pixels);
		if(mask)
			memFree(mask);
		return -1;
	}
	if(image_height > SCAN_WIDTH)
	{
		conlPrintf(LOG_ERROR,"Bitmap height must be less than %u, got %u\n",SCAN_WIDTH,image_height);
		memFree(pixels);
		if(mask)
			memFree(mask);
		return -1;
	}
	if(mask)
		mask_pass(pixels,mask);
	else
		band_pass(pixels,lo_pass,hi_pass);
	SaveImageEx(outname,image_width,image_height,pixels,space_hsv);
	memFree(pixels);
	if(mask)
		memFree(mask);
	if(fft_context != LIBV_INVALID_HANDLE)
		fftDestroyContext(fft_context);
	return 0;
}

#endif

u32_t log_two(u32_t val)
{
	u32_t count = 0;
	while((1 << count) < val)
		count++;
	return count;
}

int main(void)
{
	vhandle_t fft_context;
	framebuf_p fb;
	matrix_p c1=NULL,c2=NULL,c3=NULL;
	logStartConsoleLog();
	logUpperLevel(LOG_TRACE);
	fb = fbLoadEx("30.jpg",true,true,true);
	fft_context = fftCreateContext(log_two(fb->width),true);
	ctransFill3Matrices(fb,CSPACE_HSV,&c1,&c2,&c3);

	fftForward2D(fft_context,c1);

	matZeroRows(c1,128,fb->height-1);
	matZeroCols(c1,128,fb->width-1);

	/*
	matZeroRows(c1,0,10);
	matZeroCols(c1,0,10);
	*/
	fftInverse2D(fft_context,c1);

	// cspaceMaximizeContrast(c1,c1);
	ctransStore3Matrices(fb,CSPACE_HSV,c1,c2,c3);
	fbSaveEx(fb,"out.bmp",true);

	memFree(fb);
	memFree(c1);
	memFree(c2);
	memFree(c3);
	
	fftDestroyContext(fft_context);
	return 0;
}

