#include <v_lib.h>
#include "eval.h"

void
matZeroRows(matrix_p m,u32_t row_start,u32_t row_end)
{
	u32_t r, row_len;
	V_ASSERT(m);
	if(row_start > row_end || row_start > m->rows)
		return;
	if(row_end > m->rows)
		row_end = m->rows;
	row_len = m->cols * sizeof(vec_t);
	for(r = row_start;r < row_end;r++)
		memZero(m->cell[r],row_len);
}

void
matZeroCols(matrix_p m,u32_t col_start,u32_t col_end)
{
	V_ASSERT(m);
	if(m->rows == m->cols)
	{
		matTransposeIP(m);
		matZeroRows(m,col_start,col_end);
		matTransposeIP(m);
	}
	else
	{
		matrix_p transposed = matAlloc(m->cols,m->rows);
		matTranspose(transposed,m);
		matZeroRows(transposed,col_start,col_end);
		matTranspose(m,transposed);
		memFree(transposed);
	}
}

void
vecRunningSum(vector_p dest,vector_p src)
{
	u32_t i;
    vec_p d = dest->elements, s = src->elements;
	V_ASSERT(src->size == dest->size);
    d[0] = s[0];
    for(i = 1;i < dest->size;i++)
    	d[i] = s[i] + d[i-1];
}

void
vecLinSpaced(vector_p dest,vec_t min,vec_t max)
{
	u32_t i, n_divs;
    vec_t delta;
    n_divs = dest->size;
    delta = (max - min) / (n_divs - 1);
    dest->elements[0] = min;
    for(i = 1;i < n_divs;i++)
    	dest->elements[i] = dest->elements[i-1] + delta;
}

void
matScaleAdd(matrix_p dest,matrix_p src,vec_t scale,vec_t offset)
{
	u32_t r, c;
	V_ASSERT(dest->rows == src->rows);
	V_ASSERT(dest->cols == src->cols);
	for(r = 0;r < dest->rows;r++)
		for(c = 0;c < dest->cols;c++)
			dest->cell[r][c] = offset + src->cell[r][c] * scale;
}

